import pygame, math, camera, model, random, settings, os
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

s = math.sqrt(0.5)
arrows = dict(((K_UP, (0, 1)), (K_DOWN, (0, -1)), (K_LEFT, (-1, 0)), (K_RIGHT, (1, 0)),
    (K_HOME, (-s, s)), (K_PAGEUP, (s, s)), (K_PAGEDOWN, (s, -s)), (K_END, (-s, -s))))


def drawText(position, textString, cache={}):
    if textString not in cache:
        font = pygame.font.Font(None, 28)
        textSurface = font.render(textString, True, (255,255,255,255), (0,0,0,255)).convert_alpha()
        w, h = textSurface.get_width(), textSurface.get_height()
        cache[textString] = w, h, pygame.image.tostring(textSurface, "RGBA", True)
    w, h, textData = cache[textString]
    glRasterPos3d(*position)     
    glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, textData)

def drawGauge(position, amount, cache={}):
    level = max(min(int(amount * 50), 50), 0)
    if level not in cache:
        surf = pygame.Surface((10, 50)).convert_alpha()
        surf.fill((0, 0, 0, 0))
        pygame.draw.rect(surf, (255, 128, 0), pygame.Rect((0, 50-level), (10, level)))
        pygame.draw.rect(surf, (255, 0, 0), surf.get_rect(), 2)
        cache[level] = pygame.image.tostring(surf, "RGBA", True)
    data = cache[level]
    glRasterPos3d(*position)
    glDrawPixels(10, 50, GL_RGBA, GL_UNSIGNED_BYTE, data)


class Tower:
    model = None
    def __init__(self, p = (0, 0, 0), r = 40):
        if not self.model:
            self.model = model.Model("tower-1")
        self.r = r
        self.x, self.y, self.z = 0, 0, 0
        self.vx, self.vy = 0, 0
        self.ax, self.ay = 0, 0
        self.px, self.py, self.pz = p
        self.h = 100.
        self.n = 3.
        self.k = 2.
        self.b = 0.2
        self.q = 0.2
        self.rmax, self.topost = 0, False
        self.guys = set()
        self.neighbors = set()
        self.approaching = {}
        self.children = set()
        self.etrigger, self.etimer = 0., 0.

    def addguy(self, guy):
        self.guys.add(guy)
        guy.tower = self
        if guy == you: camera.tower = self

    def addneighbor(self, other):
        self.neighbors.add(other)
        self.approaching[other] = False

    def addforce(self, x, y, m):
        self.ax += self.q * m * x
        self.ay += self.q * m * y

    def constrain(self, (x, y, z)):
        p = math.sqrt(x ** 2 + y ** 2)
        r = self.r - 3.
        if p > r:
            return x * r / p, y * r / p, self.z
        if self.children:
            r0 = max(c.r for c in self.children) * 0.68 + 3.
            if p < r0:
                return x * r0 / p, y * r0 / p, self.z
        return x, y, self.z

    def worldcoord(self, (x, y, z)):
        return self.px + self.x + x, self.py + self.y + y, self.pz + self.z + z

    def tick(self, dt):
        for guy in self.guys:
            guy.tick(dt)
            if not guy.jumping:
                self.addforce(guy.x, guy.y, guy.m)

        self.ax += -self.k * self.x - self.b * self.vx
        self.ay += -self.k * self.y - self.b * self.vy
        self.x += self.vx * dt + 0.5 * self.ax * dt * dt
        self.y += self.vy * dt + 0.5 * self.ay * dt * dt
        self.vx += self.ax * dt
        self.vy += self.ay * dt
        self.ax, self.ay = 0, 0
        if self.guys and self.neighbors:
            self.checkjump()
        if self.etrigger:
            if math.sqrt(self.x ** 2 + self.y ** 2) > self.etrigger:
                self.etimer += dt
            else:
                self.etimer = 0.

    def checkjump(self):
        for other in self.neighbors:
            # Position of the other tower in self-centered coordinates
            x = (other.px + other.x) - (self.px + self.x)
            y = (other.py + other.y) - (self.py + self.y)
            z = (other.pz + other.z) - (self.pz + self.z)
            # Relative velocity of other tower
            vx = other.vx - self.vx
            vy = other.vy - self.vy
            app = vx * x + vy * y < 0
#            if self.approaching[other] and not app:
            if True:
                for guy in self.guys.copy():
                    gx, gy, gz = guy.x - x, guy.y - y, guy.z - z  
                    nx, ny, nz = other.constrain((gx, gy, gz))
                    dx, dy, dz = nx - gx, ny - gy, nz - gz
                    r = math.sqrt((dx ** 2 + dy ** 2) * (guy.vx ** 2 + guy.vy ** 2))
                    # Is the guy running toward that tower?
                    if not dx * guy.vx + dy * guy.vy > 0.8 * r: continue
                    # Attempt to jump to the position
                    if not guy.jumpto((dx, dy, dz)): continue
                    guy.x, guy.y, guy.z = gx, gy, gz
                    other.addguy(guy)
                    self.guys.remove(guy)
            self.approaching[other] = app


    def render(self):
        glPushMatrix()
        x = self.px + self.x
        y = self.py + self.y
        z = self.pz + self.z
        glTranslatef(x, y, z)
        glPushMatrix()
        glScalef(self.r, self.r, self.r)
        self.model.render()
        glPopMatrix()
        for guy in self.guys:
            guy.render()
        glPopMatrix()


        if you in self.guys:
            v = min(max(math.sqrt(self.x ** 2 + self.y ** 2) / 25., 0), 1.)
            creak.set_volume(v)
            if self.etimer:
                drawText((x, y, z + 10), "%.1f" % self.etimer)
            else:
                drawGauge((x, y, z + 10), v)


class TopTower(Tower):
    def __init__(self, parent, dz = 20, r = 40):
        self.parent = parent
        self.parent.children.add(self)
        Tower.__init__(self, p = (parent.px, parent.py, parent.pz + dz), r = r)
    
    def addforce(self, x, y, m):
        self.parent.addforce(x, y, m)
        
    def render(self):
        self.x, self.y, self.z = self.parent.x, self.parent.y, self.parent.z
        Tower.render(self)

class Guy:
    modelstand = None
    modelrun = None
    g = 40.
    m = 1.
    def __init__(self):
        if not self.modelstand:
            self.modelstand = model.Model("you_000001")
            self.modelrun = [model.Model("you_%06d" % (j+123)) for j in range(60)]
        self.x, self.y, self.z = 0, 0, 0
        self.vx, self.vy, self.vz = 0, 0, 0
        self.v = settings.vel
        self.tower = None
        self.fangle = 0
        self.animclock = 0
        self.walking, self.running, self.swiping = False, False, False
        self.jumping = False
        self.hurtclock = 0.

    def setv(self, (x, y)):
        if self.jumping: return
        self.vx = self.v * x * (0.5 if self.hurtclock else 1.)
        self.vy = self.v * y * (0.5 if self.hurtclock else 1.)

    def jumpto(self, (dx, dy, dz)):
        v0 = self.v
        dr = math.sqrt(dx ** 2 + dy ** 2)
        H = v0 ** 2 / self.g
        D = H ** 2 - dr ** 2 - 2 * H * dz
        if D <= 0: return False
        T = (H - math.sqrt(D)) / dr
        C = 1. / math.sqrt(1. + T**2)
        self.jt = 0.
        self.jtf = dr / (v0 * C)
        self.jvx = dx / self.jtf
        self.jvy = dy / self.jtf
        self.jvz = v0 * C * T
        self.jumping = True
        return True

    def tick(self, dt):
        if self.jumping:
            self.x += self.jvx * dt
            self.y += self.jvy * dt
            self.z += self.jvz * dt - 0.5 * self.g * dt ** 2
            self.jvz -= self.g * dt
            self.jt += dt
            if self.jt >= self.jtf: self.jumping = False
        else:
            self.x += self.vx * dt
            self.y += self.vy * dt
            if self.vx or self.vy:
                self.fangle = math.atan2(-self.vx, self.vy) * 57.3
                self.x, self.y, self.z = self.tower.constrain((self.x, self.y, self.z))
                self.running = True
            else:
                self.running = False
                self.moveclock = 0
        self.animclock += dt
        self.hurtclock = max(self.hurtclock - dt, 0.)

    def render(self):
        if self.hurtclock:
            glLightfv(GL_LIGHT0, GL_AMBIENT, (1, 0, 0, 1.0))
        glPushMatrix()
        glTranslatef(self.x, self.y, self.z)
        glRotatef(self.fangle, 0, 0, 1)
        if self.jumping:
            self.modelrun[0].render()
        elif self.swiping:
            self.modelswipe[int(self.animclock * 60) % 60].render()
        elif self.walking:
            self.modelwalk[int(self.animclock * 60) % 120].render()
        elif self.running:
            self.modelrun[int(self.animclock * 60) % 60].render()
        else:
            self.modelstand.render()
        glPopMatrix()
        if self.hurtclock:
            glLightfv(GL_LIGHT0, GL_AMBIENT, (0.2, 0.2, 0.2, 1.0))

class Guard(Guy):
    modelstand = None
    modelwalk = None
    modelrun = None
    modelswipe = None
    def __init__(self):
        if not self.modelwalk:
            self.modelstand = model.Model("guard-3_000001")
            self.modelrun = [model.Model("guard-3_%06d" % (j+123)) for j in range(60)]
            self.modelswipe = [model.Model("guard-3_%06d" % (j+184)) for j in range(60)]

            self.modelwalk = [model.Model("guard-3_%06d" % (j+2)) for j in range(120)]
        Guy.__init__(self)
        self.following = True
        self.v = random.uniform(28, 38)

    def tick(self, dt):
        if self.jumping:
            pass
        elif self.swiping:
            self.vx, self.vy = 0, 0
            if self.animclock <= 0.7 and self.animclock + dt > 0.7:
                px = self.x - 8 * math.sin(self.fangle / 57.3)
                py = self.y + 8 * math.cos(self.fangle / 57.3)
                d = math.sqrt((px - you.x) ** 2 + (py - you.y) ** 2)
                if d < 6:
                    you.hurtclock = 3.
            if self.animclock > 1.:
                self.swiping = False
        elif you.tower == self.tower and not you.jumping:
            dx, dy = you.x - self.x, you.y - self.y
            d = max(math.sqrt(dx ** 2 + dy ** 2), 0.1)
            if d < 5.:
                self.vx, self.vy = 0, 0
                self.swiping = True
                self.walking = False
                self.animclock = 0.
            else:
                if int(self.animclock * 5.) != int((self.animclock + dt) * 5.):
                    self.vx, self.vy = dx * self.v / d, dy * self.v / d
                    self.walking = False
        else:
            if random.random() * 2. < dt:
                theta = random.random() * 2*math.pi
                self.vx, self.vy = 10. * math.cos(theta), 10 * math.sin(theta)
                self.walking = True
            elif random.random() * 2. < dt:
                self.vx, self.vy = 0, 0
                self.walking = False
        Guy.tick(self, dt)

    def jumpto(self, p):
        if settings.levelname == "level3":
            return Guy.jumpto(self, p)
        return False


pygame.init()
pygame.mixer.init()

if settings.music:
    pygame.mixer.music.load(os.path.join("data", "aquellos-dias-dogy-fu.ogg"))
    pygame.mixer.music.play(-1)
    pygame.mixer.music.set_volume(settings.mvolume)

creak = pygame.mixer.Sound(os.path.join("data", "creak.wav"))
creak.play(0)
creak.set_volume(0)

scr = pygame.display.set_mode(settings.resolution, DOUBLEBUF | OPENGL | (FULLSCREEN if settings.fullscreen else 0))
sx, sy = scr.get_width(), scr.get_height()

pygame.display.set_caption("The Teetering Towers of San Gimignano")

glutInit()
glLightfv(GL_LIGHT0, GL_POSITION,  (100, 100, 100, 1.))
glLightfv(GL_LIGHT0, GL_AMBIENT, (0.2, 0.2, 0.2, 1.0))
glLightfv(GL_LIGHT0, GL_DIFFUSE, (0.5, 0.5, 0.5, 1.0))
glEnable(GL_LIGHT0)
glEnable(GL_LIGHTING)
glEnable(GL_COLOR_MATERIAL)
glEnable(GL_DEPTH_TEST)
glDepthFunc(GL_LESS)
glEnableClientState(GL_VERTEX_ARRAY)
glEnableClientState(GL_NORMAL_ARRAY)
glShadeModel(GL_SMOOTH)
glEnable(GL_BLEND)
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
glEnable(GL_TEXTURE_2D)
glEnable(GL_RESCALE_NORMAL)
glClearColor(0.4, 0.7, 0.9, 1.0)
glEnable(GL_FOG)
glFogi(GL_FOG_MODE, GL_LINEAR)
glFogfv(GL_FOG_COLOR, (0.4, 0.7, 0.9, 1.0))


glMatrixMode(GL_PROJECTION)
glLoadIdentity()
gluPerspective(40.0, sx/float(sy), 1, 1000.0)
glEnable(GL_DEPTH_TEST)

clock = pygame.time.Clock()

quitting = False

fnumber = 0

while not quitting:

    you = Guy()

    towers = {}
    tneigh = {}
    triggers = {}
    for line in open(os.path.join("data", "levels.txt"), "r"):
        w = line.strip().split(" ")
        if not w: continue
        if w[0] != settings.levelname: continue
        if w[1] == "start":
            stower = w[2]
        elif w[1] == "trigger":
            triggers[towers[w[2]]] = " ".join(w[3:])
        elif w[1] == "etrigger":
            towers[w[2]].etrigger = float(w[3])
        elif w[1] == "guard":
            towers[w[2]].addguy(Guard())
        else:
            tname = w[1]
            if "-t" in tname:
                dz, r = w[2:4]
                tneigh[tname] = w[4:]
                towers[tname] = TopTower(towers[tname[:-2]], float(dz), float(r))
            else:
                x, y, z, r = w[2:6]
                tneigh[tname] = w[6:]
                towers[tname] = Tower((float(x), float(y), float(z)), float(r))

    for tname, neighs in tneigh.items():
        for neigh in neighs:
            if neigh in towers:
                towers[tname].addneighbor(towers[neigh])

    towers[stower].addguy(you)

    pygame.mouse.set_visible(False)
    pygame.event.set_grab(True)

    playing = True
    fpstime = 1.
    firstframe = True
    while playing:
        dt = clock.tick() * 0.001
        if firstframe:
            dt = 0.001
            firstframe = False
        for e in pygame.event.get():
            if e.type == KEYDOWN and e.key in (K_ESCAPE, K_q):
                playing = False
                quitting = True
            elif e.type == KEYDOWN and e.key == K_1:
                settings.levelname = "level1"
                playing = False
            elif e.type == KEYDOWN and e.key == K_2:
                settings.levelname = "level2"
                playing = False
            elif e.type == KEYDOWN and e.key == K_3:
                settings.levelname = "level3"
                playing = False
            elif e.type == MOUSEMOTION:
                camera.move(e.rel)
            elif e.type == MOUSEBUTTONDOWN and e.button == 4:
                camera.zoomin()
            elif e.type == MOUSEBUTTONDOWN and e.button == 5:
                camera.zoomout()
            elif e.type == KEYDOWN and e.key == K_INSERT:
                camera.zoomin(6.)
            elif e.type == KEYDOWN and e.key == K_DELETE:
                camera.zoomout(6.)

        m = pygame.key.get_pressed()
        kx, ky = 0, 0
        for key, (keyx, keyy) in arrows.items():
            if m[key]:
                kx += keyx
                ky += keyy
        if settings.levelname == "tutorial": ky = 0

        if kx or ky:
            absk = math.sqrt(kx ** 2 + ky ** 2)
            kx /= absk
            ky /= absk

            you.setv(camera.transform(kx, ky))
        else:
            you.setv((0, 0))


        for tower in towers.values():
            tower.tick(dt)

        glMatrixMode(GL_MODELVIEW)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()

        camera.look(dt)
        for tower in towers.values():
            tower.render()


        if settings.showfps:
            fpstime += dt
            if fpstime >= 1:
                fps = clock.get_fps()
                fpstime -= 1

        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        glOrtho(0, 1, 0, 1, -1, 1)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        text = "%s: %s fps" % (settings.levelname, int(fps)) if settings.showfps else settings.levelname
        drawText((0.05, 0.95, 0), text)

        if you.tower in triggers:
            drawText((0.05, 0.05, 0), triggers[you.tower])

        glMatrixMode(GL_PROJECTION)
        glPopMatrix()


        if settings.levelname == "level2" and "tower11" in towers and you.tower == towers["tower11"] and not you.jumping:
            settings.levelname = "level3"
            playing = False

        if settings.levelname == "tutorial" and "tower9" in towers and you.tower == towers["tower9"] and not you.jumping:
            settings.levelname = "level1"
            playing = False

        if settings.levelname == "level1" and "tower9" in towers and towers["tower9"].etimer > 10:
            settings.levelname = "level2"
            playing = False



        pygame.display.flip()


    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    pygame.display.flip()

